#include "Rinterface.h"

#include <qfile.h>
#include <qmessagebox.h>

#include <unistd.h>

// OSX specific include for determining bundle location
#if defined(Q_OS_MACX)
#include <CoreFoundation/CoreFoundation.h>
#endif

// the constructor which optionally starts the server as well
Rinterface::Rinterface(bool startServer) {
  if (startServer == true) {
    status = establishConnection();
  }
  else {
    status = NOTSTARTED;
  }
  socketEstablished = false;
  socket = new Q3Socket();
  connect(socket, SIGNAL(hostFound()), this, SLOT(socketHostFound()));
  connect(socket, SIGNAL(error(int)), this, SLOT(socketErrorFound(int)));
  connect(socket, SIGNAL(readyRead()), this, SLOT(socketMessage()));
  connect(socket, SIGNAL(connected()), this, SLOT(socketConnected()));

  pipe = new Q3Process( );
  connect(pipe, SIGNAL(readyReadStderr()), this, SLOT(processErrorMessage()));

  progress = new Q3ProgressDialog("R","Cancel", 2, 0, "R progress", FALSE);
  progressBar = new Q3ProgressBar();
  progressBar->setTotalSteps(2);
  //progress->setBar(progressBar);
  connect(this, SIGNAL(startProgressDialog(QString)), 
	  this, SLOT(progressStarted(QString)));
  connect(this, SIGNAL(stopProgressDialog()), this, SLOT(progressFinished()));
  

  socketErrorCount = 0;
  maxSocketErrors = 25;
  socketPort = 6012;
  socketHost = "127.0.0.1";
}

// establishes the connections with the R-server pipe
Rinterface::connectionStatus Rinterface::establishConnection() {
  connectionStatus returnStatus;
  // only do something if the connection has not already been established
  if (status != OK) {

    QString serverCmd = "R-server";

    // Mac OSX specific stuff - assume the R-server command is inside
    // the application bundle
#if defined(Q_OS_MACX)
    CFURLRef pluginRef = CFBundleCopyBundleURL(CFBundleGetMainBundle());
    CFStringRef macPath = CFURLCopyFileSystemPath(pluginRef, 
						  kCFURLPOSIXPathStyle);
    const char *pathPtr = CFStringGetCStringPtr(macPath, 
						CFStringGetSystemEncoding());
    qDebug("Path = %s", pathPtr);
    serverCmd = pathPtr;
    serverCmd.append("/Contents/MacOS/R-server");
    CFRelease(pluginRef);
    CFRelease(macPath);
#endif


    pipe->addArgument(serverCmd);
    pipe->addArgument("/tmp/R-log.txt");
    // start the pipe
    if (! pipe->start()) {
      returnStatus = CONNECTIONERROR;
    }
    else {
      // now try to start the socket
      status = OK;
      cout << "Establishing socket" << endl;

      // the R command to start a socket server. BUG: should replace port
      // and host with class definitions socketPort and socketHost
      sendCommand(tr("con1 <- socketConnection(port = 6012, server=TRUE)"));

      // the qt command to connect to a socket
      startSocket();

      sendCommand("writeLines(\"testing\", con1)");
      cout << "End of socket stuff" << endl;
      returnStatus = OK;
    }
  }
  status = returnStatus;
  return(returnStatus);
}

// when an error is found, print the error number and retry
void Rinterface::socketErrorFound(int errorNum) {
  cout << "Error: " << errorNum << endl; 
  if (socketErrorCount < maxSocketErrors) {
    startSocket();
    socketErrorCount++;
    cout << "Retrying socket connection ..." << endl;
  }
  else {
    cout << "Final socket connection attempt failed ..." << endl;
  }
  sleep(1);
}

void Rinterface::sendRSocketMessage(QString message) {
  QString cmd;
  cmd.sprintf("writeLines(\"%s\", con1)", message.ascii());
  sendCommand(cmd);
}

void Rinterface::socketConnected() {
  cout << "connected" << endl; 
  socketEstablished = true;
  emit connected();
}

// this is where socket messages are processed
void Rinterface::socketMessage() {
  QString message;
  while(socket->canReadLine()) {
    message.append( socket->readLine() );
  }
  cout << "Message: " << qPrintable(message);
  // some messages are handled here, others passed along for whatver
  // is using this interface.
  if (message.startsWith("RDONE:")) {
    emit stopProgressDialog();
  }
  else {
    emit newSocketData(message);
  }
}

// reads the error stream from the process and displays a message box with it
void Rinterface::processErrorMessage() {
  cout << "received error message" << endl;
  QString errorMessage;
  while (pipe->canReadLineStderr()) {
    errorMessage.append( pipe->readLineStderr() );
  }
  QMessageBox::critical( 0, tr("Process Error"),
                         errorMessage, tr("Quit") );
}

// sends a command to the R pipe
void Rinterface::sendCommand(QString command, bool progressDialog,
			     QString progressString) {
  if (status == OK) {
    command.append("\n");
    cout << "R COMMAND: " << qPrintable(command);
    pipe->writeToStdin(command);
    if (progressDialog) {
      emit startProgressDialog(progressString);
    }
  }
  else {
    //throw noConnectionException();
    //cout << "after throw" << endl;
  }

}

// BUG: gets confused when there is more than one event that is being watched.
// BUG #2: idiotic updating of progress - it is always at 50%
void Rinterface::progressStarted(QString progressString) {
  progress->setLabelText(progressString);
  progress->reset();
  progress->setProgress(1);
  sendRSocketMessage("RDONE:");
}

void Rinterface::progressFinished() {
  progress->setProgress(2);
}

// load a data file
bool Rinterface::loadDataFile(QString datafile) {
  QString loadCommand;
  loadCommand.sprintf( "load(\"%s\")", datafile.ascii() );
  sendCommand(loadCommand, true, "loading R data");
}

void Rinterface::startSocket(){
  socket->connectToHost(socketHost, socketPort);
}


// the destructor
Rinterface::~Rinterface() { 
  sendCommand(tr("QUIT"));
}
